'use strict';
// プログラムとかよくわかんねえ！！！な場合でも、ここだけ見て数値を変えれば色々変えられます
// その他の部分は魔法陣かなんかだと考えて無視してください
const charInterval = 50 // 文字表示間隔（ミリ秒）
const talkInterval =15 // トーク表示間隔（秒）

const address = {
	dict: "ghost/master/dic_talk.txt",
	surfaceFolder: "shell/master/",
	surfaceTxt: "shell/master/surfaces.txt",
	replaceTxt: "ghost/master/replace.txt",
	replaceAfterTxt: "ghost/master/replace_after.txt",
	satoriconf: "ghost/master/satori_conf.txt",
	touch: "ghost/master/dic_menu.txt",
}
// 起動ボタンのスタイルシート
const bootButtonStyle = {
	width: 'fit-content',
	border: 'solid black 1.5px',
	'border-radius': '5px',
	background: 'linear-gradient(wheat, white)',
	padding: '8px',
	left: '0px',
	position: 'fixed',
	bottom: '0px',
	'z-index': 100
}
// ボタン上に表示される起動状態、左から●起動中、◆ロード中、★動作中
const displayedStates = [ '●おためし表示', '◆loading...', '★非表示にする', '★再表示する' ]
// バルーンのスタイルシート
const baloonStyle = [{
  'font-family': "メイリオ, MS P明朝, MS 明朝, MS ゴシック, sans-serif",
	border: 'solid black 1.5px',
	'border-radius': '5px',
	width: '326px',
	height: '129px',
	'line-height': 1.3,
	right: '195px',
	bottom: '135px',
	position: 'fixed',
	display: 'none',
	background: 'linear-gradient(wheat, white)',
	padding: '8px',
	'white-space': 'pre-wrap',
	'z-index': 99,
	'box-sizing': 'unset'
}]
// △▼△△▼△△▼△△▼△△▼△△▼△△▼△△▼
// 　　　　　　　　　以下魔法陣
// △▼△△▼△△▼△△▼△△▼△△▼△△▼△△▼
// ゴースト本体の画像canvas
const surfaceCanvasStyle = [{
	display: 'none',
	position: 'fixed',
	bottom: '0px',
	right: 0,
	height: 'auto',
	'z-index': 99
}]
// surface.txt用のクラス
class SurfaceElement {
	constructor(imageAddress, offsetX, offsetY) {
	  this.imageAddress = imageAddress;
	  this.offsetX = offsetX;
	  this.offsetY = offsetY;
	}
}
// トーク読み上げ用
// unread -> readedに一文字づつ移していくようなイメージ
// 標準で二個入り
class TalkText {
	constructor(text){
		this.readed = ['', '']
		this.unread = text
	}
	clear() {
		this.readed = ['', '']
		this.unread = ''
	}
	// 一文字読み上げる
	read (targetID) {
		this.readed[targetID] += this.unread.substr(0, 1);
		this.unread = this.unread.substr(1);
	}
	// 指定された文字までのテキストを切り出す
	// 返り値にその文字は含まない（ただしその文字を含めて読み上げた扱いとなる）
	readToChar(startIndex, char) {
		const endIndex = this.unread.indexOf (char);
		const result = this.unread.substr(startIndex, endIndex-1);
		this.unread = this.unread.substr(endIndex + 1);
		return result
	}
	readTo(startIndex = 0, endIndex = 1) {
		const result = this.unread.substr(startIndex, endIndex);
		this.unread = this.unread.substr(endIndex);
		return result
	}
	// 指定分の文字を読み飛ばす
	skip (value) {
		this.unread = this.unread.substring(value);
	}
	// 次の文字を確認
	seeNextChar (startIndex) {
		return this.unread.substr(startIndex, 1);
	}
}
// △▼△▼△▼△▼△▼△▼△▼△▼△▼△▼△▼
// トーク関連（バルーンもここ）
// △▼△▼△▼△▼△▼△▼△▼△▼△▼△▼△▼
class TalkController {
	constructor() {
		this.intervalID = 0
		this.talkTarget = 1
		this.talkText
		this.baloon_HTMLs = getHTMLElements('baloon_')
		this.talk_in_baloon_HTMLs = getHTMLElements('talk_in_baloon_')
	}
	init () {
		this.resetSurface()
	}
	bootPrepare() {
		this.#setTextSize()
		this.#placeBaloonsToSideOfCanvas()
		this.startRandomTalk()
	}
	talk(text) {
		clearInterval(this.intervalID);

		this.#resetSurfaceAndBaloon();
		this.talkText = new TalkText(text)
		this.talkTarget = 1

		this.intervalID = setInterval(()=> {
			this.#readAloud();
		}, charInterval);
	}
	startRandomTalk() {
		const randomTalk = ghostData.talkDict[getRandomInt(0, ghostData.talkDict.length)];
		this.talk(randomTalk);
	}
	startRandomTouchTalk(touchPoint) {
		const randomTouchTalk = this.#getRandomTouchTalk(touchPoint);
		if (randomTouchTalk == null) return
		this.talk(randomTouchTalk);
	}
	#getRandomTouchTalk(key) {
		const talklist = ghostData.touchDict[key];
		if (checkIsUndefined(talklist))
			return;
		const random = getRandomInt(0, talklist.length);
		return talklist[random];
	}
	// 指定時間停止、その後再開
	async #waitTalk(sleepTime) {
		let intervalID_WhenWaitStart = this.intervalID;
		clearInterval(this.intervalID);
		await sleep(sleepTime);
		
		// 別の場所から割り込みでトークが始まっていれば中断
		if (this.intervalID != intervalID_WhenWaitStart)
			return false;
		
		this.intervalID = setInterval(()=> {
			this.#readAloud();
		}, charInterval);
		return true;
	}
	#reflectTalk(targetId) {
		this.talk_in_baloon_HTMLs[targetId].textContent = this.talkText.readed[targetId]
	}
	// 読み上げ実行、talkstoreより一文字づつ読み出し反映していく
	// 特殊な記号に対しては必要な範囲を切り出して使用
	async #readAloud() {
		let nextChar = this.talkText.seeNextChar()
		if(checkIsUndefined(nextChar))
			return;
		switch (nextChar) {
			// ：＝トークの対象切り替え
			case '：':
				this.#changeTalkTarget();
				this.talkText.skip(1)
				break;
			// \＝色々あるので別メソッドに
			case '\\':
				// awaitを挟むと割り込まれる可能性があるため、割り込まれた場合は中断の処理
				const success = await this.#readMetaChar();
				if (!success) return
				break;
			// （）＝サーフェス変更、変数としての用途はされない（読み込み時に削除）
			case '（':
				let surfaceNum = this.talkText.readToChar(1, '）')
				// 全角 -> 半角変換
				surfaceNum = zenkakuNumToHankakuNum(surfaceNum);
				this.#showSurface(this.talkTarget, surfaceNum);
				break;
			default:
				this.talkText.read(this.talkTarget)
		}
		
		this.#showBaloon(this.talkTarget);
		this.#scrollTalkInBaloon();
		this.#reflectTalk(this.talkTarget)
	
		// talkstoreすべて読了時、次のトークを予約
		if (nextChar == '')
			await this.#reserveNextTalk();
		
	}
	// 新しいトークを準備
	async #reserveNextTalk() {
		let intervalID_WhenReserveStart = this.intervalID;
		clearInterval(this.intervalID);

		// 喋り間隔の半分経過でバルーンの表示オフ
		await sleep(talkInterval / 2);

		// 割り込まれていれば中断
		if (intervalID_WhenReserveStart != this.intervalID)
			return;
		
		// サーフェスリセット+バルーン非表示で再待機
		this.#resetSurfaceAndBaloon();

		await sleep(talkInterval / 2);

		// 割り込まれれば中断
		if (intervalID_WhenReserveStart != this.intervalID)
			return;

		this.startRandomTalk();

	}
	// '\'を読み込んだ場合の処理、現在はウェイトのみ
	async #readMetaChar() {
		// '\'の次の文字より処理を決定
		const readChar = this.talkText.seeNextChar(1)
		
		switch(readChar) {
			case 'w': // ウェイト
				const waitTime = parseInt(this.talkText.seeNextChar(2));
				if (checkIsUndefined(waitTime))
					break;
				// 割り込みがなく正常終了出来ていればtalkから読み込んだ分を削除
				const isFinished = await this.#waitTalk(0.05 * waitTime);
				if (isFinished) 
					this.talkText.readTo(0, 3)
				return isFinished;
			// とりあえず怪しげな半角英数字&記号を省くという雑な処置
			// 基本的にトーク辞書読み込み時に弾かれているはずだが念のため設定
				default:
					const jpIndex = this.talkText.unread.search(/[^ -~]/);
					
					if(jpIndex == -1)
						this.talkText.unread = "";
					else
						this.talkText.readTo(0, jpIndex-1);
					return true;
		}
		this.talkText.readTo(0, 1)
	}
	// バルーン内のトークをスクロールさせる
	#scrollTalkInBaloon () {
		for (let i = 0; i < baloonStyle.length; i++) {
			// 折り返しは改行コードでなく自動で行われているため、一行の距離はここで図る	
			const baloonHeight = parseInt(baloonStyle[i].height);
			// 行数は内部の値と外部に反映されるものが一文字分ズレるので、ここで直接計算する
			let oneLineLimit = parseInt (baloonStyle[i].width) / this.textSize;
	
			// トークの行数計測
			const splitedTalk = this.talkText.readed[i].split(/\r\n|\n/);
			let talkTextLine = 0;
	
			// 改行コードで分割したトークを数える（折り返しを考慮）
			for (let i = 0; i < splitedTalk.length; i++) {
				const split = splitedTalk[i];
				// 最後の行が改行コードのみ場合は計算に入れない
				if (split.length == 0 && i == splitedTalk.length-1)
					break
				// 1行の文字数 / 最大文字数の切り捨て後が1以上ならその分行数が増える
				const lineFloat = split.length / oneLineLimit + 1;
				talkTextLine += Math.trunc(lineFloat);
			}
			const talkHeight = talkTextLine * this.textSize * baloonStyle[i]['line-height'];

			if (baloonHeight < talkHeight) {
				// splitの一行目が折り返し不要の場合はそちら分の文字を差し引く
				const newLineCodeIndex = splitedTalk[0].length;
				if (newLineCodeIndex < oneLineLimit) {
					oneLineLimit = newLineCodeIndex; 
					this.talkText.readed[i] = this.talkText.readed[i].replace(/\r\n|\n/, '');// 行端の改行コードを削除
				}
				this.talkText.readed[i] = this.talkText.readed[i].substring(oneLineLimit);
			}
		}
	}
	// トークの表示バルーン、サーフェスの反映対象の切り替え
	// もし3人以上に対応する日が来ればここを変えよう
	#changeTalkTarget() {
		this.talkTarget = this.talkTarget == 0 ? 1 : 0;
	}
	// テキストの大きさを算出、保存（バルーン内スクロールとかに使う）
	#setTextSize(){
		var e = document.getElementById("ruler");
		var width = e.clientWidth;
		this.textSize = width;
	}
	#placeBaloonsToSideOfCanvas () {
		baloonStyle[0].right = canvasController.surfaceCanvases[0].canvases[0].clientWidth + 'px';
		Object.assign(this.baloon_HTMLs[0].style, baloonStyle[0])

		// 2人以上の場合はまず1人目の配置の値をコピーし、一部を置換する
		if (canvasController.surfaceCanvases.length > 1){
			baloonStyle[1] = JSON.parse(JSON.stringify((baloonStyle[0])));
			baloonStyle[1].right = canvasController.surfaceCanvases[0].width;

			// baloon1はバルーン0の下の余白に詰め込む
			baloonStyle[1].bottom = 0;
			baloonStyle[1].height = 
				parseInt(baloonStyle[0].bottom) - (parseInt(baloonStyle[0].padding) * 2) - 3; // -3はborder線の太さ*2、ほぼ変更されないだろう値
			Object.assign(this.baloon_HTMLs[1].style, baloonStyle[1])
		}
	}
	// ＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃
	// ＃＃○サーフェス・バルーン表示関連＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃
	// ＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃
	#showSurface(id, surfaceId) {
		const surfaceName = "surface" + calcSurfaceNum(surfaceId);
		const surfaceElements = ghostData.surfaceElements[surfaceName];
	
		if (checkIsUndefined(surfaceElements))
			return;
		
		canvasController.surfaceCanvases[id].draw(ghostData.surfaceImages, surfaceElements);
		canvasController.nowSurface[id] = surfaceName;
	}
	#showBaloon(i) {
		if(this.baloon_HTMLs[i].style.display == 'none')
			this.baloon_HTMLs[i].style.display = 'block';
	}
	#resetSurfaceAndBaloon() {
		for (let i = 0; i < this.baloon_HTMLs.length; i++) {
				this.baloon_HTMLs[i].style.display = 'none';
		}
		this.resetSurface();
	}
	resetSurface() {
		for (let i = 0; i < canvasController.surfaceCanvases.length; i++) {
			this.#showSurface(i, ghostData.defaultSurface[i]);
		}
	}
}
// レイヤ機能の付いたCanvasを擬似的に表現する
// 要はCanvasを重ねる
class LayerCanvas {
	constructor(canvas) {
		this.canvases = new Array(canvas);
		this.ctxes = new Array(canvas.getContext('2d'));
	}
	push(canvas) {
		this.canvases.push(canvas);
		this.ctxes.push(canvas.getContext('2d'));
	}
	getImageData(index) {
		return this.ctxes[index].getImageData(0, 0, this.width, this.height);
	}
	createImageData(index) {
		return this.ctxes[index].createImageData(this.width, this.height);
	}
	drawImage(layerId, image) {
		this.canvases[layerId].width = image.width;
		this.canvases[layerId].height = image.height;
		this.ctxes[layerId].drawImage(image, 0, 0, this.width, this.height);
	}
	// surfaceimagesを渡すのは無理やりな気もする
	draw(images, elements) {
		// 0枚目のキャンバスに描画
		const element_0 = elements[0];
		const image_0 = images[element_0.imageAddress];
		if (image_0 == undefined)
			return;

		this.canvases[0].width = image_0.width;
		this.canvases[0].height = image_0.height;
		this.ctxes[0].putImageData(image_0, element_0.offsetX, element_0.offsetY);

		// 残りを同様に描画、ただしキャンバスの大きさは0枚目に合わせる
		for (let i = 1; i < elements.length; i++) {
			const iElement = elements[i];
			const iImage = images[iElement.imageAddress];
			
			if (iImage == undefined)
				continue;

			this.canvases[i].width = this.width;
			this.canvases[i].height = this.height;
			this.ctxes[i].putImageData(iImage, iElement.offsetX, iElement.offsetY);
		}

		// 使用されなかったキャンバス上の画像を削除
		for (let i = elements.length; i < this.canvases.length; i++) {
			const clearImage = this.createImageData(i);
			this.ctxes[i].putImageData(clearImage, 0, 0)
		}
	}
	get width() {
		return this.canvases[0].width;
	}
	get height() {
		return this.canvases[0].height;
	}
}
// 辞書からトークを引っ張ってくるときに使います
// header = ＊行、talk = {}内
class TalkDefine {
	constructor(header, talkstore) {
		this.header = header;
		this.talk   = talkstore;
	}
}
// Square & Polygon2Dはsurface当たり判定用のクラス
class Square {
	// 開始点(左上xy)と幅、高さ
	constructor(x, y, xEnd, yEnd) {
		this.x = x;
		this.y = y;
		this.xEnd = xEnd;
		this.yEnd = yEnd;
	}
	isIn(x, y) {
		const isInX = x > this.x && x < this.xEnd;
		const isInY = y > this.y && y < this.yEnd;
		return isInX && isInY;
	}	
}
class Vertex {
	constructor(x, y) {
		this.x = x
		this.y = y
	}
}
class Polygon2D {
	constructor(x, y) {
		this.polygon = new Array()
		this.push(x, y)
	}
	push(x, y) {
		this.polygon.push(new Vertex(Number(x), Number(y)))
	}
	// AIが8割方書いてくれました。私が数日掛けたやつより確実に高性能です。えらい。
	isIn(x, y) {
		let inside = false;
		for (let i = 0, j = this.polygon.length - 1; i < this.polygon.length; j = i++) {
			let xi = this.polygon[i].x, yi = this.polygon[i].y;
			let xj = this.polygon[j].x, yj = this.polygon[j].y;
	
			let intersect = ((yi > y) != (yj > y))
				&& (x < (xj - xi) * (y - yi) / (yj - yi) + xi);
			if (intersect) inside = !inside;
		}
	
		return inside;
	}
}
// 起動ボタン関連の処理
class BootButton {
	constructor () {
		this.displayedState = '' // ボタン表示
		this.button_HTML = document.getElementById('boot-button')
	}
	init (state, style, clickEvent) {
		Object.assign(this.button_HTML.style, style)
		this.button_HTML.addEventListener('click', clickEvent)
		this.changeStateDisplay(state)
	}
	changeStateDisplay(state) {
		this.#changeStateAndReflectDisplay(state)
	}
	#changeStateAndReflectDisplay(state) {
		const stateString = displayedStates[state]
		this.button_HTML.textContent = stateString
	}
}
// ゴーストの読み込みに関わるクラス
// サーフェスもこちら
// 割と密結合というか、トーク辞書をサーフェス読み込みに利用したりしますので、機能で切り分けるのは難しそう
class GhostData {
	constructor() {
		this.initValues 
		this.replaceTexts 
		this.talkDict 
		this.touchDict
		this.defaultSurface
	}
	async load() {
		await this.#setAddress()
		await this.#loadDicts()
		await this.#prepareSurfaces();
	}
	async #setAddress() {
		const configText = await this.#loadText('ushs_config.txt');
		const splitedConfig = configText.split(/\r\n|\n/);
		const targetGhostID = this.#getTargetGhostID()
		
		// ushs_config.txtの内容を適用、設定idはonbootbuttonclick時のhtmlからの引数
		for (let i = 0; i < splitedConfig.length; i++) {
			const iConfig = splitedConfig[i];
			const configIndex = iConfig.indexOf('◆USHS設定');
			if (configIndex != 0) 
				continue;
	
			const id = parseInt(iConfig.substring(7));
			if (id != targetGhostID)
				continue;
				
			const ghostFolderIndex = splitedConfig.indexOf('#ゴースト本体フォルダ', i);
			const ghostFolder = splitedConfig[ghostFolderIndex+1];
			address.folder = ghostFolder + "/";
			const talkFileIndex = splitedConfig.indexOf('#ランダムトーク辞書ファイル', ghostFolderIndex);
			const talkFile = splitedConfig[talkFileIndex+1];
			address.dict = address.folder + "ghost/master/" + talkFile;
			const touchFileIndex = splitedConfig.indexOf('#触り反応辞書ファイル', talkFileIndex);
			const touchFile = splitedConfig[touchFileIndex+1];
			address.touch = address.folder + "ghost/master/" + touchFile;
			break;
		}
		address.surfaceFolder = address.folder + address.surfaceFolder;
		address.surfaceTxt = address.folder + address.surfaceTxt;
		address.replaceTxt = address.folder + address.replaceTxt;
		address.replaceAfterTxt = address.folder + address.replaceAfterTxt;
		address.satoriconf = address.folder + address.satoriconf;
	}
	#getTargetGhostID() {
		const bootButtonHTML = document.getElementById('boot-button')
		return bootButtonHTML.dataset.ghost_id
	}
	async #loadDicts () {
		this.initValues 	= await this.#loadInitValues()
		this.replaceTexts = await this.#loadReplace();
		this.talkDict 		= await this.#makeTalks();
		this.touchDict 		= await this.#makeTouchs();
		this.#setDefaultSurface()
	}
	// ロード時、Shift_JIS -> UTF-8に自動変換される。元の文字コードに注意
	async #loadText(address) {
		const response = await fetch(address)
		const arrayBuffer = await response.arrayBuffer()
		const textDecoder = new TextDecoder("Shift_JIS")
		return textDecoder.decode(arrayBuffer);
	}
	// 辞書からランダムトークを生成
	// ※isDeniedTalkに引っかかるトークは生成されない
	async #makeTalks() {
		let text = await this.#loadText(address.dict);
		let splitedText = this.#convertTextToTalkDefines('＊', '＊', text);
		let talks = [];
		for (let i = 0; i < splitedText.length; i++) {
			let talkstore = splitedText[i].talk;
			talkstore = this.#applyTalkOptions(talkstore);
			if (this.#isDeniedTalk(splitedText[i].header, talkstore))
				continue;
			else
				talks.push(talkstore);
		}
		return talks;
	}
	// 辞書から触り反応を以下略
	async #makeTouchs() {
		const text = await this.#loadText(address.touch);
		let talkDefines = this.#convertTextToTalkDefines('＊', '＊', text);
		let touchs = {};
		for (let i = 0; i < talkDefines.length; i++) {
			const header = talkDefines[i].header;
			let talk = talkDefines[i].talk;
			// 0または1から始まる(例：0headなでられ)トークのみ対象（例外あるが今はこう）
			const firstChar = header.replace(/＊/, '').substring(0, 1);
			if (firstChar != '0' && firstChar != '1') 
				continue;
			// replace.txt, satori_conf.txtを適用
			talk = this.#applyTalkOptions(talk);
			if (this.#isDeniedTalk(header, talk))
				continue;
			
			if (checkIsUndefined(touchs [header]))
				touchs [header] = new Array(talk);
			else
				touchs [header].push(talk);
		}
		return touchs;
	}
	// 許可できないトークか判定
	// 現状は表示に条件を含むもの（'* （親愛度）＞３'など）
	// 特殊変数を含むもの（'(if~)'など）
	// 別トークに飛ばすもの'＞'など）
	#isDeniedTalk(header, talk) {
		const isDeniedHeader = (/\t/).test(header);
		const isDeniedTalk = (/(（[^０１２３４５６７８９])|＞|_/).test(talk);

		return isDeniedHeader || isDeniedTalk;
	}
	#removeDisallows(talk) {
		const removedVariableColumn = talk.replace(/^＄.*((\r\n|\n|\r)|$)/mg, '');
		const removedEscape = removedVariableColumn.replace(/φ/g, '');
		return removedEscape;
	}
	async #loadReplace() {
		let replaceText =  await this.#loadText(address.replaceTxt);
			replaceText += await this.#loadText(address.replaceAfterTxt);
		const splitedReplace = replaceText.split(/\r\n|\n/);
		return splitedReplace;
	}
	async #loadInitValues() {
		const initValueText = await this.#loadText(address.satoriconf);
		const splitedInitValue = initValueText.split(/\r\n|\n/);
		return splitedInitValue;
	}
	// satori_conf.txt & replace.txtの内容を適用
	// 変数を含むトークへの初期値割り当て（初期値がなければトーク自体を削除）
	#applyTalkOptions(text) {
		if(checkIsUndefined(text))
			return text;
		text = this.#applyInitValue(text);
		text = this.#applyReplaceTxt(text);
		// 不要な記号などを削除
		text = this.#removeDisallows(text);
		return text;
	}
	// satori_conf
	#applyInitValue (text) {
		// 行単位に分割 -> '＄'が文頭の場合に置換処理
		for (let i = 0; i < this.initValues.length; i++) {
			let firstchar = this.initValues[i].substring (0, 1);
			if (firstchar != '＄')
				continue;
			
			let splitedInitValue = this.initValues[i].split(/\t/); 
			if (checkIsUndefined (splitedInitValue) || splitedInitValue.length != 2)
				continue;
			// 初期値をトーク中を検索するために置換
			// ※例：＄ユーザ名 -> （ユーザ名）
			splitedInitValue[0] = '（' + splitedInitValue[0].replace('＄', '') + '）';
			splitedInitValue[0] = splitedInitValue[0].replace(/[.*+?^=!:${}()|[\]\/\\]/g, '\\$&'); // メタ文字をエスケープ
			splitedInitValue[0] = new RegExp (splitedInitValue[0], 'g');
			text = text.replace(splitedInitValue[0], splitedInitValue[1]);
		}
		return text;
	}
	// replace.textの内容を適用
	#applyReplaceTxt(text) {
		for (let i = 0; i < this.replaceTexts.length; i++) {
			let splitedReplaces = this.replaceTexts[i].split(/\t/); 
			if (checkIsUndefined (splitedReplaces) || splitedReplaces.length != 2)
				continue;
			
			// 文章中にreplace.txtに対応したものがあれば置き換え
			splitedReplaces[0] = splitedReplaces[0].replace(/[.*+?^=!:${}()|[\]\/\\]/g, '\\$&'); // メタ文字をエスケープ
			splitedReplaces[0] = new RegExp (splitedReplaces[0], 'g');
			text = text.replace(splitedReplaces[0], splitedReplaces[1]);
		}
		return text;
	}
	// ひとつ下のメソッドを'{}'で囲まれたtextに使用するためのメソッド（消し残しの'{'を削除する以外に特殊な措置はない）
	#convertBracesToTalkDefines(headerMarker, text) {
		const talkDefines = this.#convertTextToTalkDefines(headerMarker, '}', text);
		return talkDefines.map(x => { x.talk.replace(/{/, ''); return x; });
	}
	// splitMarkerで区切られたテキストを
	// header: headerMarkerから始まる行, talkstore: それ以降...に分割します
	#convertTextToTalkDefines(headerMarker, splitMarker, text) {
		let talkDefines = [];
		let splitedTexts = text.split(splitMarker);
		
		 // 最初のsplitMarkerまでの余分な情報を削除（余分な情報がなければ無視）
		if (splitedTexts[0].indexOf('surface') == -1)
			splitedTexts.shift();
		
		// headerMarkerとsplitMarkerが同一の場合、
		// split時にheaderMarkerのための情報が消失するため補完
		const isSameMarker = headerMarker == splitMarker;
		if(isSameMarker) { splitedTexts = splitedTexts.map(x => { return headerMarker + x }) }

		const headerRegExp = new RegExp('^' + headerMarker + '.*', 'm');
		for(let i = 0; i < splitedTexts.length; i++) {
			let split = splitedTexts[i];
			let header = splitedTexts[i].match(headerRegExp);
			const talkRegExp = new RegExp ('(?<=' + header + '(\r\n|\n|\r)).*', 's');
			let talkstore = split.match(talkRegExp)
			if (talkstore == null || header == null)
				continue;
			header = header[0];
			// コメントアウトをスパッと削除 
			talkstore = talkstore[0].replace(/\/\/.*|([(#＃].*)/gm, '');
			
			// 無駄改行とかを削除... /replace(/^(\r\n|\n|\r)/, '')で同じ結果が得られるはずだがダメでした（理由不明）
			const splitedTalks = talkstore.split(/\r\n|\r|\n/);
			let adjustedTalk = "";
			splitedTalks.forEach((column, i) => { 
				if(column.length == 0) return;
				if(adjustedTalk.length != 0) adjustedTalk += "\r\n";
				adjustedTalk += column;
				if(i < splitedTalks.length-1) adjustedTalk;
			});
			talkDefines.push(new TalkDefine(header, adjustedTalk));
		}
	
		return talkDefines;
	}
	// ＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃
	// ＃＃○画像読み込み関連＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃
	// ＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃
	#setDefaultSurface() {
		this.defaultSurface = [0, 100]

		const target = '＄デフォルトサーフェス';
		this.initValues.forEach(value => {
			if (value.match(target)) {
				const values = value.replace(target, '').split(/\t/);
				this.defaultSurface[values[0]] = values[1];
			}
		});
	}
	// shell/masterフォルダの情報を元にsurfaceを生成
	// 当たり判定のセットなども行われる
	async #prepareSurfaces() {
		// サーフェス番号にパディング（先頭に0を足す）が必要か算出
		await this.#calcValidDigit();
		// images(pngファイル)とelements(surfaces.txt)を抽出
		const surfaceDefines = await this.#getSurfacesDefines();
		this.surfaceImages = await this.#loadSurfaceImages(surfaceDefines);
		 // definesからelementsとcollisionsに切り分け、適用
		const elements_collisions = this.#definesToElements_Collisions(surfaceDefines);
		this.collisionList = elements_collisions.collisions;
		this.surfaceElements = elements_collisions.elements;
	}
	// Surface.txtからサーフェス定義（Define）を抽出
	async #getSurfacesDefines(){
		// 将来的に使用するサーフェス番号のみをロード
		let useSurfaceNumbers = this.#getUseSurfaceNumbers();
		
		const surfacesText = await this.#loadText(address.surfaceTxt);
		const talkDefines = this.#convertBracesToTalkDefines('surface', surfacesText);
		let surfaceDefines = this.#getUseNumberDefines(useSurfaceNumbers, {}, talkDefines);
		
		surfaceDefines = this.#element0Complement(surfaceDefines);
		surfaceDefines = this.#convertAnimationsToElements(surfaceDefines, talkDefines);
		return surfaceDefines;
	}
	#getUseNumberDefines(useSurfaceNumbers, surfaceDefines, talkDefines) {
		// ②①の番号に対応したtalkDifineを取得、surfaceDefinesに登録
		for(let i = 0; i < talkDefines.length; i++) {
			// surface0 {...}の'surface0'部分
			const surfaceHeader = talkDefines[i].header;
			let surfaceSettings = talkDefines[i].talk.match(/(?<=(^|\n|\r\n|\r))(element|collision|animation).*/g)
			if (surfaceSettings == null)
				continue;
	
			// 対象となるサーフェス番号を取得
			const targetSurfaces = this.#getTargetSurfaces(surfaceHeader);
	
			// それに対応するDefineのみをsurfaceDefinesに書き出す
			targetSurfaces.forEach(target => {
				// 対象となるSurface番号が全トーク中に使用されるかチェック、不要ならカット
				const surfaceNum = target.replace(/surface/, '');
				const paddedNum = calcSurfaceNum(surfaceNum);
				const regExp = new RegExp('(?<=（)' + paddedNum + '(?<!）)');
				if(!regExp.test(useSurfaceNumbers))
					return;
	
				target = target.replace (surfaceNum, paddedNum)
	
				if (checkIsUndefined(surfaceDefines [target]) || surfaceDefines [target] == null)
					surfaceDefines [target] = JSON.parse(JSON.stringify(surfaceSettings)); // ディープコピー
				else
					surfaceDefines [target] = surfaceDefines [target].concat (surfaceSettings);
			});
		}
		return surfaceDefines;
	}
	// element0要素がなければ補完します
	#element0Complement (surfaceDefines) {
		Object.keys(surfaceDefines).forEach(key => {
			let iElements = surfaceDefines[key];
			let isHasElement0 = false;
			
			// element要素がなければ補完
			if (iElements == null) {
				surfaceDefines[key] = new Array('element0,overlay,' + key + '.png,0,0');
				return;
			}
			// element要素はあるがelement0がないか確認、あればreturn
			for (let i = 0; i < iElements.length; i++) {
				isHasElement0 = (/element0/).test(iElements[i]);
				if(isHasElement0)
					break;
			}
			if(isHasElement0)
				return;
			// なければ追加
			else
				iElements.push('element0,overlay,' + key + '.png,0,0');
		});
		return surfaceDefines;
	}
	// だいぶ長いです相当長いです実際長いです
	// 読んで字のごとくsurfaceDifine内のanimation(bind)……着せ替え表示用の定義
	// これを通常のelementの形に変換します
	#convertAnimationsToElements (surfaceDefines, talkDefines) {
		let foolproof = 0;
		let loadedSurfaceNums = "";
		// 作業が必要なくなるまでループ（foolproofは万一用）
		while (foolproof < 100) {
			const startLength = Object.keys(surfaceDefines).length;
			let useSurfaceNums = "";

			// ①surfaceDefinesに登録されているanimation(bind)から必要なサーフェス番号を抽出
			Object.keys(surfaceDefines).forEach (key => {
				const define = surfaceDefines[key];
				
				for (let i = 0; i < define.length; i++) {
					// animation...internal,bind行からanimationのidを抽出
					const targetId = define[i].match (/(?<=animation)[0-9]*(?=.interval,bind)/);
					if (targetId == null) continue;
					// それ以降の行からidに対するadd行を抽出、internal,bindからaddは連続したものと扱う（それ以外の行が挟まると正常動作しない）
					for (let j = i + 1; j < define.length; j++) {
						const matchRegExp = new RegExp ('(?<=animation' + targetId[0] +'.pattern[0-9]*,add,).*');
						const addColumn = define[j].match(matchRegExp)
						if(addColumn == null) continue;
						const animationDefine = addColumn[0].split(/,/);
						const surfaceId = '（' + calcSurfaceNum(animationDefine[0]) + '）';
						if (!loadedSurfaceNums.includes(surfaceId))
							useSurfaceNums += surfaceId;
					}
				}
			});
			// ②①の番号に対応したtalkDifineを取得、surfaceDefinesに登録
			surfaceDefines = this.#getUseNumberDefines(useSurfaceNums, surfaceDefines, talkDefines);
			const endLength = Object.keys(surfaceDefines).length;
			
			// ③②の結果、新たにanimation(bind)を有するelementsが増えている可能性があるため①へ戻る
			// ただし、このループも何も登録されていない場合はその可能性がないため終了
			if(startLength == endLength)
				break;
			loadedSurfaceNums += useSurfaceNums;
			foolproof ++;
		}
		
		// ④ここまでの結果でsurfaceDefinesには必要なサーフェス定義がすべてある
		// ここでanimation(bind)定義から対応するelementを取得（無駄なanimation定義は削除検討、今のところさほど害がないため保留？）
		Object.keys(surfaceDefines).forEach (key => {
			const defineColumns = surfaceDefines[key];
			let appends = [];
			
			for (let i = 0; i < defineColumns.length; i++) {
				// animation...internal,bind行からanimationのidを抽出
				const defineColumn = defineColumns[i];
				const targetId = defineColumn.match (/(?<=animation)[0-9]*(?=.interval,bind)/);
				if (targetId == null)
					continue;
	
				// それ以降の行からidに対するadd行を抽出、internal,bindからaddは連続したものと扱う（それ以外の行が挟まると正常動作しない）
				for (let j = i + 1; j < defineColumns.length; j++) {
					const matchRegExp = new RegExp ('(?<=animation' + targetId[0] +'.pattern[0-9]*,add,).*');
					const addColumn = defineColumns[j].match(matchRegExp)
					if(addColumn == null)
						continue;
					const animationsElements = addColumn[0].split(/,/);
					const surfaceKey = 'surface' + calcSurfaceNum(animationsElements[0]);
					const offsetX = animationsElements[2];
					const offsetY = animationsElements[3];
					const offsetedElements = this.#elementOffsetAtDefine(surfaceDefines[surfaceKey], offsetX, offsetY);
					
					if (appends.includes(offsetedElements))
						continue;
					else
						appends.push (offsetedElements);
				}
			}
			if (appends.length == 0) 
				return;
			appends.forEach(append => {
				if (surfaceDefines[key].includes(append)) 
					return;
				else
					surfaceDefines[key] = surfaceDefines[key].concat(append)
			});
		});
		return surfaceDefines;
	}
	// define内のelementのoffsetX,Yの値にx,yを足したものを返します
	// animation(bind)に指定されているoffsetXYを加算する用
	#elementOffsetAtDefine(define, x, y) {
		let offsetedDefines = JSON.parse(JSON.stringify(define)); // ディープコピー
		
		for (let i = 0; i < offsetedDefines.length; i++)  {
			let element = offsetedDefines[i]
			const fOffsetXRegExp = new RegExp('(?<=element[0-9]*,.*,.*.png,).*(?=,)')
			const fOffsetYRegExp = new RegExp('(?<=element[0-9]*,.*,.*.png,.*,[0-9]*).*')
			let fOffsetX = element.match(fOffsetXRegExp);
			let fOffsetY = element.match(fOffsetYRegExp);
			if (fOffsetX == null | fOffsetY == null)
				continue;
			fOffsetX = parseInt(fOffsetX[0]);
			fOffsetY = parseInt(fOffsetY[0]);
			fOffsetX += x;
			fOffsetY += y;
			element = element.replace(fOffsetXRegExp, fOffsetX)
			element = element.replace(fOffsetYRegExp, fOffsetY)
			offsetedDefines[i] = element;
		};
		return offsetedDefines;
	}
	// サーフェスpng画像をロードし、透過ImageDataに変換する
	async #loadSurfaceImages (elements) {
		 let surfaceImages = {}; // ImageData
		// 必要な画像のみをロードする
		const needSurfacePngs = this.#getNeedSurfaceName(elements);
		 for(let i = 0; i < needSurfacePngs.length; i++) {
			 const needsurfacePng = needSurfacePngs[i];
			 // ImageData取得のため一度Canvas上に描画
			 let image = await this.#loadImage(address.surfaceFolder + needsurfacePng).catch(e => {});
			 if (typeof image === "undefined") continue;
			 canvasController.surfaceCanvases[0].drawImage (0, image);
		
			 // 透過済みか確認、済んでいればそのまま、いなければ透過してsurfacesにpush
			 const imageData = canvasController.surfaceCanvases[0].getImageData(0);
			 const isTransparentedImage = this.#checkIsTransparentedImage(imageData);
			 if (isTransparentedImage) {
				 surfaceImages [needsurfacePng] = imageData;
			 } 
			else {
				 const transparentedImage = this.#makeTransparentedImage(); // 描画中のImageDataを透過済みに変換
				 surfaceImages [needsurfacePng] = transparentedImage;
			}
		 }
		 return surfaceImages;
	}
	async #loadImage (src) {
		 return new Promise((resolve, reject) => {
			 const img = new Image();
			 img.onload = () => resolve(img);
			 img.onerror = (e) => reject(e);
			 img.src = src;
		 });
	}
	// surface0.pngをロード失敗した際に呼ばれる
	// surface00.png ~ surface00000000.pngまでロードを施行し、成功した桁数を記録
	// 以降、サーフェス番号指定時はその桁数に自動でパディングされる
	async #calcValidDigit() {
		for (let digit = 0; digit < 8; digit++) {
			const loadResult = await this.#loadImage(address.surfaceFolder + 'surface' + getPaddingString(digit) + "0" + '.png').catch(e => {});
			const isLoaded = !checkIsUndefined (loadResult);
			if (isLoaded) {
				this.surfacePadDigit = digit + 1; 
				return;
			}
		}
	}
	// elementsに記載されているpng画像名を抽出
	#getNeedSurfaceName(elements) {
		 let needPngs = []
		 Object.keys(elements).forEach(key => {
			const surfaceDefines = elements[key];
			 surfaceDefines.forEach(defineColumn => {
				const png = defineColumn.match(/(?<=,)[^,]*png/);
				if (png == null)
					return;
				 const isUndiscovered = !needPngs.includes(png[0]);
				 if (isUndiscovered)
					 needPngs.push(png[0]);
			 });
		 });
	
		 return needPngs;
	}
	// トーク辞書から実際に使われているサーフェス番号を取得
	#getUseSurfaceNumbers() { 
		let matchedSurface = "";
		const surfaceNumRegExp = /（[０１２３４５６７８９]*）/g;
		this.talkDict.forEach(x => {
			const matched = x.match(surfaceNumRegExp);
			if(matched == null) return;
			matched.forEach(m => { if (!matchedSurface.includes(m)) matchedSurface += m; })
		});
		// 可読性が悪いがやっていることは上記（talkdict.forEach）とほぼ同一
		Object.keys(this.touchDict).forEach(key => {
			let matched = [];
			this.touchDict[key].forEach(fTarget => { 
				const fMatched = fTarget.match(surfaceNumRegExp);
				if (fMatched == null) return;
				matched = matched.concat(fMatched);
			});
			
			matched.forEach(m => { if (!matchedSurface.includes(m)) matchedSurface += m; })
		});

		matchedSurface = zenkakuNumToHankakuNum(matchedSurface);
		matchedSurface = this.#complementDefaultSurface(matchedSurface);
		matchedSurface = replaceSurfaceNums(matchedSurface);

		return matchedSurface;
	}
	// matchedSurfaceにデフォルトサーフェス番号がない場合に補完します
	#complementDefaultSurface(matchedSurface) {
		for (let i = 0; i <  this.defaultSurface.length; i++) {
			const defaultSurfaceNum = '（' + this.defaultSurface[i] + '）'
			const defaultRegExp = new RegExp(defaultSurfaceNum);
			if (!(defaultRegExp).test(matchedSurface))
				matchedSurface += defaultSurfaceNum;
		}
		return matchedSurface;
	}
	// surfaceHeader(例：surface0, 119, 10-100...)から対象となるsurfaceの配列を返します
	#getTargetSurfaces(surfaceHeader) {
		// .appendは現状取得時に取り除き元の数値と同列に扱っています
		surfaceHeader = surfaceHeader.replace(/(.append|\s)/g, '');
		const isContainComma = (/,/).test(surfaceHeader);
		const surfaceHeaders = isContainComma ? surfaceHeader.replace(/,/g, ',surface').split(',') : [surfaceHeader]; 
		let targetSurfaces = [];
		surfaceHeaders.forEach(header => {
			const isRange = (/-/).test(header);
			if (isRange) {
				const range = header.replace(/surface/, '').split('-')
				for (let i = range[0]; i <= range[1]; i++) {
					const surface = 'surface' + i;
					targetSurfaces.push(surface);
				}
			} else {
				targetSurfaces.push(header);
			}
		});
		return targetSurfaces;
	}
	// definesをelementsとcollisionsに切り分ける
	#definesToElements_Collisions(surfaceDefines) {
		let _surfaceElements = {};
		let surfaceCollisions = {};
		 // definesからそれぞれ必要なデータを取り出し格納
		 Object.keys(surfaceDefines).forEach(key => {
			const elements = this.#getSurfaceElements(surfaceDefines[key]);
			const collisons = this.#getSurfaceCollisionPoints(surfaceDefines[key]);
			surfaceCollisions[key] = collisons;
			if (elements.length == 0)
				return;
			 _surfaceElements [key] = elements;
		});
		return { elements: _surfaceElements, collisions: surfaceCollisions};
	}
	#getSurfaceElements(surfaceDefines) {
		 let _surfaceElements = [];
		 
		 for (let i = 0; i < surfaceDefines.length; i++) {
			// element0,overlay,surface0.png,0,0...を','で切り分け
			 const defineElements = surfaceDefines[i].split(',');
			 // elementが適正な形ではない（省略を考慮する必要？）
			 if (defineElements.length != 5)
				 continue;
			 const iImageAddress = defineElements[2];
			 const iOffsetX	 	= defineElements[3]
			 const iOffsetY		= defineElements[4]
			 _surfaceElements.push(new SurfaceElement(iImageAddress, iOffsetX, iOffsetY));
		 }
		 return _surfaceElements;
	}
	// ＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃
	// ＃＃○当たり判定設定関連＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃
	// ＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃
	// 複数行のcollisionから当たり判定部分を抽出
	#getSurfaceCollisionPoints(defineArray) {
		 let collisionPoints = {};
		 for (let i = 0; i < defineArray.length; i++) {
			const defineColumn = defineArray[i];
			const isCollisionColumn = (/collision/).test(defineColumn);
			 if (!isCollisionColumn)
				 continue;
			const defineElements = defineColumn.split(',');
			const isEx = (/ex/).test(defineColumn);
			let reactPoint;
			let collision;
			 if(isEx) {
				 reactPoint = defineElements[1];
				 collision  = this.#makeExCollision(defineElements);
			 } 
			 else {
				reactPoint = defineElements[5];
				collision = this.#makeNormalCollision(defineElements);
			 }
			 
			if (reactPoint in collisionPoints)
				collisionPoints[reactPoint].push(collision);
			else
				collisionPoints[reactPoint] = new Array(collision);
		 }
		 return collisionPoints;
	}
	// 行がEX(CollisionEx)の場合
	// （例：collisionex5,hara,polygon,54,220,58,258,125,262,128,240,129,216,88,215,66,217）
	#makeExCollision(defineElements) {
		const x = defineElements[3];
		const y = defineElements[4];
		// ExPolygon
		if (defineElements[2] == 'polygon') {
			let polygon2D = new Polygon2D(x, y);
		 
			for (let j = 5; j < defineElements.length - 1; j+=2) {
				const jx = defineElements[j];
				const jy = defineElements[j+1];
				polygon2D.push(jx, jy);
			}
			
			return polygon2D;
		} 
		 // ExSquare
		else {
			const x = defineElements[3];
			const y = defineElements[4];
			const xEnd = defineElements[5];
			const yEnd = defineElements[6];
			const square = new Square(x, y, xEnd, yEnd);
			return square;
		}
	}
	// 行がnormal(Collision)の場合
	// （例：collision6,38,286,56,307,hand）
	#makeNormalCollision(defineElements) {
		 const x			= defineElements[1];
		 const y			= defineElements[2];
		 const width	= defineElements[3];
		 const height	= defineElements[4];
		 const square	= new Square(x, y, width, height);
		 return square;
	}
	// ＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃
	// ＃＃○透過画像作成関連＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃
	// ＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃
	// 指定したcanvasから透過画像を生成
	#makeTransparentedImage() {
		let sourceImage = canvasController.surfaceCanvases[0].getImageData(0);
		// 左上ドットの色を取得、比較用Json生成
		const transparentColor = JSON.stringify(this.#getDotColor(sourceImage, 0));
		// 左上ドットに透明度が設定されていれば透過処理不要と判断
		const isAlreadyTransparented = sourceImage.data[3] == 0;
		if (isAlreadyTransparented) return;
		// 元画像のドット一つ一つの色を取得、transparentedにコピー
		// コピーの際、その色がtransparentColorと同色なら透過（alpha=0に）
		let transparentedImage = canvasController.surfaceCanvases[0].createImageData(0);
		for (let i = 0; i < sourceImage.data.length; i += 4) {
			const iColor = this.#getDotColor(sourceImage, i);
			const isNeedTransparent = JSON.stringify (iColor) == transparentColor;
			iColor[3] = isNeedTransparent ? 0 : 255; // alpha
		
			transparentedImage = this.#setColor (transparentedImage, i, iColor);
		}
		return transparentedImage;
	}
	// ImageData
	#getDotColor(imageData, i) {
		 const dotColor = [
			 imageData.data[i],		// r
			 imageData.data[i + 1],	// g
			 imageData.data[i + 2],	// b
			 imageData.data[i + 3] 	// a
		 ];
		 return dotColor;
	}
	// imageDataの指定地点iにcolorをセット
	#setColor (imageData, i, color) {
		 imageData.data[i]   = color[0];
		 imageData.data[i+1] = color[1];
		 imageData.data[i+2] = color[2];
		 imageData.data[i+3] = color[3];
		 return imageData;
	}
	// Imageが透過済みかチェック
	#checkIsTransparentedImage (imageData){
		// 左上ドットのalphaが255以外ならば透過イメージとみなす
		const dot = this.#getDotColor(imageData, 0);
		return dot[3] < 255;
	}
}
class CanvasController {
	constructor () {
		this.nowSurface = []
		this.canvas_HTMLs = []
		this.surfaceCanvases = []
		this.#searchCanvasHTMLs()
	}
	// 既にhtml上に存在するCanvasを登録、現状は最大3つとする
	#searchCanvasHTMLs() {
		this.canvas_HTMLs = getHTMLElements('ushs-s')
		for (let i = 0; i < this.canvas_HTMLs.length; i++) {
			this.surfaceCanvases[i] = new LayerCanvas(this.canvas_HTMLs[i]);
		}
	}
	#placeSecondCanvas() {
		surfaceCanvasStyle[1] = JSON.parse(JSON.stringify(surfaceCanvasStyle[0]));
		surfaceCanvasStyle[1].right =	parseInt(this.surfaceCanvases[0].width) + parseInt(baloonStyle[0].width) + (parseInt(baloonStyle[0].padding) * 2);
	}
	// もろもろ初期化
	// ゴーストデータの読み込みなどが済んだ前提で呼ばれています
	init () {
		this.#makeCanvasLayers()
	}
	bootPrepare() {
		this.#placeSecondCanvas()
		this.#setCanvasLayersStyle()
		this.#addLisnerToCanvas()
		this.#showCanvasWhenBoot()
	}
	// LayeredCanvasのStyleをセット
	// 基本的に0番目0のコピーだがpointer-eventの無効化とz-indexだけは元の値+にする必要がある
	#setCanvasLayersStyle() {
		for (let i = 0; i < this.surfaceCanvases.length; i++) {
			// 0の値のコピーを設定
			let styleString = "pointer-events: none;";
			const baseStyles = surfaceCanvasStyle[i];
			Object.keys(baseStyles).forEach(key => {
				styleString += key + ":" + baseStyles[key] + ";";
			});
			let zIndex = parseInt(styleString.match(/(?<=z-index:)[^;]*/));
			
			for (let j = 1; j < this.surfaceCanvases[0].canvases.length; j++) {
				const iZIndex = zIndex + i;
				const iStyle = styleString.replace(/(?<=z-index:)[^;]*/, iZIndex);
				this.surfaceCanvases[i].canvases[j].setAttribute('style',iStyle);
			}
		}
	}
	// surfaceCanvasに重ねる新しいCanvasを生成
	#makeCanvasLayers() {
		const surfaceElements = ghostData.surfaceElements
		let maxCount = 0;

		// surfaceElements内のsurfaceElement数＝サーフェス一つが使用する画像数
		// その最大値が必要となるレイヤの数
		Object.keys(surfaceElements).forEach(key => {
			const keysElements = surfaceElements[key];
			maxCount = Math.max(keysElements.length, maxCount)
		});
		
		// レイヤを生成、どのサーフェスをどのキャンバスに表示するか制限はできないので、キャンバス全てに最大数のレイヤを設定
		for (let i = 0; i < this.surfaceCanvases.length; i++) {
			for(let j = 1; j < maxCount; j++) {
				const createdCanvas = document.createElement('canvas');
				document.getElementById('ushs').append(createdCanvas);
				this.surfaceCanvases[i].push(createdCanvas);
			}
		}
	}
	// 初期値では非表示になっているCanvasを表示、起動時一度のみ呼ばれる
	#showCanvasWhenBoot() {
		for (let i = 0; i < surfaceCanvasStyle.length; i++) {
			surfaceCanvasStyle[i].display = 'block';
		}

		// 全てのCanvasのStyle"display"を"block"に変更
		for (let i = 0; i < this.surfaceCanvases.length; i++) {
			Object.assign(this.surfaceCanvases[i].canvases[0].style, surfaceCanvasStyle[i])

			for (let j = 1; j < this.surfaceCanvases[i].canvases.length; j++) {
				const target = this.surfaceCanvases[i].canvases[j];
				const targetStyle = target.getAttribute('style')
				if(targetStyle == null) continue;
				const blockedStyle = targetStyle.replace(/(?<=display:)[^;]*/, 'block');
				target.setAttribute('style', blockedStyle)
			}
		}
	}
	#addLisnerToCanvas() {
		for (let i = 0; i < this.canvas_HTMLs.length; i++) {
			this.canvas_HTMLs[i].addEventListener('pointerdown', reactionController.onmouseclick.bind(reactionController, i))
			this.canvas_HTMLs[i].addEventListener('pointermove', reactionController.onmousemove.bind(reactionController, i))
		}
	}
}
// ＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃
// ＃＃○さわり反応関連＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃
// ＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃
class ReactionController {
	constructor() {
		this.brushing = ''				// なで反応用、現在なでている部位
		this.brushingTime = 0			// なで反応用、なで時間の計測用
		this.brushingGrace = 0		// なで以下略、ターゲットから外れてもなで継続扱いの猶予（ピクセル）
		this.brushingGraceMax = 8 // 猶予の最大値
		this.brushingNow = false
	}
	#getCollisionedTargets(id, e) {
		const result = []
		
		const surfaceName = canvasController.nowSurface[id];
		const collisionsOnSurface = ghostData.collisionList[surfaceName];
		if (collisionsOnSurface == null) { return; }

		// 当たり判定リストを全検査
		// クリック位置が判定内ならつつきトーク起動
		const collisionKeys = Object.keys(collisionsOnSurface);
		for (let i = 0; i < collisionKeys.length; i++) {
			const key = collisionKeys[i];
			const targets = collisionsOnSurface[key];
			if(!this.#isCollision(targets, e))
				continue;
			
			result.push(key)
		}
		return result
	}
	onmouseclick(id, e) {
		// 仕様上起動前でも呼べてしまうためここで処理（moveも同様）
		if(nowBeforeBoot())
			return;
		
		const collisionedTargets = this.#getCollisionedTargets(id, e)
		if(checkIsUndefined(collisionedTargets))
			return;

		for (let i = 0; i < collisionedTargets.length; i++) {
				const dictKey = "＊" + id + collisionedTargets[i] + "つつかれ";
				if(checkIsUndefined (ghostData.touchDict[dictKey]))
					continue;
				
				talkController.startRandomTouchTalk(dictKey);
				break;
		}
	}
	onmousemove(id, e)  {
		// 排他制御
		if (this.brushingNow || nowBeforeBoot())
			return;
		
		this.brushingNow = true;

		const collisionedTargets = this.#getCollisionedTargets(id, e)
		if (checkIsUndefined(collisionedTargets)) {
			this.brushingNow = false;
			return
		}

		// カーソルの状態変化（さわれることを示す）、addEventListnerでonmousemoveで発生するイベントは2種登録できないため、ここで
		this.#changeCursorStyle(id, collisionedTargets)

		// スタート時のなで継続時間を記録
		let brushingTimeAtStart = this.brushingTime;

		// 登録されている当たり判定リストを愚直に全検査
		// 現在のマウスカーソル位置が判定内……を一定時間（このメソッドが呼ばれた回数）維持で撫で成立とみなす
		// 維持といっても呼ばれるのはマウスカーソルが動いた時なので置きっぱなしではダメ
		// onmouseclickの検査工程と似ているが、処理が色々違うので別に書いています
		for (let i = 0; i < collisionedTargets.length; i++) {
			const isTarget = this.brushing == collisionedTargets[i];

			// なで継続中＆60ピクセル（要調整？）以上同じkeyを撫でている場合は撫でトーク
			if (isTarget && this.brushingTime > 60) {
				const dictKey =  "＊"  + id + collisionedTargets[i] + "なでられ";

				// talkControllerにトークを予約、失敗した場合はリセットしない
				talkController.startRandomTouchTalk(dictKey);
				this.#resetBlushingPoint();
				break;
			} else if (isTarget) {
				this.brushingTime++;
				break;
			}
			
			// なで継続中でない場合、どこかをなでていればそこを登録
			if (this.brushing == '') {
				this.#memoryNewBrushingPoint(collisionedTargets[i]);
				this.brushingTime++;
				break;
			}
		}
		// なで継続中にも関わらずbrushingTimeが増えてない = 失敗扱い
		const isNotKeeping = this.brushingTime == brushingTimeAtStart
		const isWithinGrace = this.brushingGrace > 0;
		// なで継続失敗、Grace猶予切れ
		if (isNotKeeping && isWithinGrace)
			this.#resetBlushingPoint();
		// 猶予はまだある
		if (isNotKeeping && !isWithinGrace)
		this.brushingGrace--;
		
		this.brushingNow = false;
	}
	// 現在地が当たり判定を持つ場合、カーソルを変化
	#changeCursorStyle(id, collisionedTargets) {
		const isAnywhereCollision = collisionedTargets.length > 0
		const cursorState = isAnywhereCollision ? 'grab' : 'auto'
		canvasController.canvas_HTMLs[id].style.cursor = cursorState
	}
	#isCollision(targets, e) {
		for (let i = 0; i < targets.length; i++) {
			if (targets[i].isIn(e.offsetX, e.offsetY))
				return true;
		}
		return false;
	}
	#memoryNewBrushingPoint(pointKey) {
		this.#resetBlushingPoint();
		this.brushing = pointKey;
	}
	#resetBlushingPoint() {
		this.brushingGrace = this.brushingGraceMax;
		this.brushingTime = 0;
		this.brushing = '';
	}
}
// ▼△▼△▼△▼△▼△▼△▼△▼△▼△▼△▼△▼△▼△▼△▼△
// ▼ここから本体
// ▼△▼△▼△▼△▼△▼△▼△▼△▼△▼△▼△▼△▼△▼△▼△
// 起動状態 0 = 起動前, 1 = 起動プロセス中, 2 = 動作中
// 画面表示は上部"displayedState"に
let appState = 0

const ghostData = new GhostData()
const talkController = new TalkController()
const canvasController = new CanvasController()
const reactionController = new ReactionController()

// 初期設定
const bootButton = new BootButton()
bootButton.init(appState, bootButtonStyle, onbootbuttonclick)

// 開始ボタン押下時
async function onbootbuttonclick(e) {
	const shouldSuspend = isNeedSuspend()
	if(shouldSuspend) {
		suspendExecution();
		return
	}

	changeState(1)

	await ghostData.load()

	// 初期化①、ロードしたデータを元に
	canvasController.init()
	talkController.init()

	// 初期化②、initで設定した情報を相互に利用する
	canvasController.bootPrepare()
	talkController.bootPrepare()

	changeState(2)
}
function isNeedSuspend() {
	return nowWaitingBoot() || nowBoot() || nowResume();
}
// 押下ボタン処理中断時に指定処理を行う
function suspendExecution() {
	if (nowWaitingBoot()) {
		return;
	}
	// 起動中に押下：非表示に変更
	if (nowBoot()) { 
		suspendUSHS();
    changeState(3)
		return;
	}
	// 非表示中に押下：表示する
  if (nowResume()) {
		resumeUSHS();
    changeState(2)
		return;
  }
}
function suspendUSHS() {
  document.getElementById('ushs').style.display = 'none'
}
function resumeUSHS() {
  document.getElementById('ushs').style.display = 'block'
}
function changeState(value) {
	appState = value
	bootButton.changeStateDisplay(value)
}
// state = 0 | 1 == 起動前、起動中はBeforeBoot
function nowBeforeBoot() {
	return appState == 0 || appState == 1;
}
// state = 1 == 起動中
function nowWaitingBoot() {
	return appState == 1;
}
// state = 2 == 実行中
function nowBoot() {
	return appState == 2;
}
// state = 3 == 中断中
function nowResume() {
	return appState == 3;
}
// ＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃
// ＃＃○その他雑多ごった＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃
// ＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃＃
async function sleep(sleepTime) {
  const _sleep = (second) => new Promise(resolve => setTimeout(resolve, second * 1000))
  await _sleep(sleepTime);
}
function getPaddingString (digit) {
	if(digit == 0)
		return "";
	let padString = ""; 
	for (let i = 0; i < digit; i++) { padString += "0" }
	return padString;
}
function replaceSurfaceNums(nums) {
	const paddingRegExp = new RegExp('0*([0-9]{' + ghostData.surfacePadDigit + ',})', 'g');
	return nums.replace(/([0-9]+)/g, getPaddingString (ghostData.surfacePadDigit) + '$1').replace(paddingRegExp, '$1');
}
function calcSurfaceNum (num) {
	if (ghostData.surfacePadDigit == 0)
		return num;
	else {
		const paddingRegExp = new RegExp('0*([0-9]{' + ghostData.surfacePadDigit + ',})');
		return num.replace(/([0-9]+)/, getPaddingString (ghostData.surfacePadDigit) + '$1').replace(paddingRegExp, '$1');
	}
}
function checkIsUndefined (data){
	return typeof data === "undefined";
}
function zenkakuNumToHankakuNum(num) {
	return num.replace(/[０-９．]/g, function(s) {
		return String.fromCharCode(s.charCodeAt(0) - 65248);
	});
}
function getRandomInt(min, max) {
	min = Math.ceil(min);
	max = Math.floor(max);
	return Math.floor(Math.random() * (max - min) + min); //The maximum is exclusive and the minimum is inclusive
}
function getHTMLElements(id) {
	let result = []
	for (let i = 0; i < 3; i ++) {
		const HTMLElement = document.getElementById(id + i)

		if (HTMLElement != null)
			result.push(HTMLElement)
	}
	return result
}